import os
import logging
from typing import Any, Optional, Type, TypeVar

# External Packages
from google import genai
from google.genai import types
from pydantic import BaseModel

# Self Defined Packages
from ..._util import CreatorRole, ChatCompletionConfig, MessageBlock, ResponseMode
from ..._core import Core, ImageInterpreter
from .base import GeminiCore

T = TypeVar("T", bound=BaseModel)
logger = logging.getLogger(__name__)


class GMN_StructuredOutput_Core(Core, GeminiCore, ImageInterpreter):
    """SO LLM"""

    def __init__(
        self,
        system_prompt: str,
        config: ChatCompletionConfig,
    ):
        Core.__init__(self, system_prompt, config)
        GeminiCore.__init__(self, config.name)
        self.profile = self.build_profile(model_name=config.name)

    def custom_config(
        self,
        max_output_tokens: int,
        response_mode: Optional[ResponseMode] = None,
        response_format: Optional[Type[T]] = None,
    ) -> types.GenerateContentConfig:
        """Adapter function.

        Transform custom ChatCompletionConfig -> types.GenerationContentConfig
        """
        config = types.GenerateContentConfig(
            system_instruction=self.system_prompt,
            temperature=self.config.temperature,
            max_output_tokens=max_output_tokens,
            # frequency_penalty=0.5,
        )
        if response_mode == ResponseMode.SO:
            config.response_mime_type = "application/json"
            config.response_schema = response_format
        elif response_mode == ResponseMode.JSON:
            config.response_mime_type = "application/json"
        return config

    def validate(
        self, response_mode: Optional[ResponseMode], response_format: Optional[Type[T]]
    ) -> None:
        if response_mode:
            if not isinstance(response_mode, ResponseMode):
                raise TypeError(
                    f"Expect mode to be an instance of 'ResponseMode', but got '{type(response_mode).__name__}'."
                )
            if response_mode is response_mode.SO:
                if response_format is None:
                    raise TypeError(
                        "Expect format to be a subclass of 'BaseModel', but got 'NoneType'."
                    )
                if not issubclass(response_format, BaseModel):
                    raise TypeError(
                        f"Expect format to be a subclass of 'BaseModel', but got '{type(response_format).__name__}'."
                    )

    def run(
        self, query: str, context: list[MessageBlock | dict] | None, **kwargs
    ) -> list[MessageBlock | dict]:
        """
        Synchronously run the LLM model with the given query and context.

        Args:
            query (str): The query to be processed by the LLM model.
            context (list[MessageBlock | dict] | None): The context to be used for the LLM model.
            **kwargs: Additional keyword arguments.
                * `filepath` (str | None): Path to the image file.
                * `mode` (ResponseMode | None): Ouput mode.
                * `format` (BaseModel | None): Output structure.

        Returns:
            list[MessageBlock | dict]: The list of messages generated by the LLM model.

        Notes:
        * Single-Turn Execution.
        """
        response_mode: Optional[ResponseMode] = kwargs.get("mode", ResponseMode.DEFAULT)
        response_format: Optional[Type[T]] = kwargs.get("format")  # type: ignore
        self.validate(response_mode, response_format)  # Raise an exception if invalid

        msgs: list[types.Content] = []
        output: list[MessageBlock | dict] = []

        if context:
            for ctx in context:
                msgs.append(
                    types.Content(
                        role=ctx["role"],
                        parts=[types.Part.from_text(text=ctx["content"])],
                    )
                )
        parts = []
        filepath: str | None = kwargs.get("filepath", None)
        if filepath:
            with open(filepath, "rb") as f:
                data = f.read()
            ext = os.path.splitext(filepath)[-1][1:]
            ext = "jpeg" if ext == "jpg" else ext
            parts.append(types.Part.from_bytes(data=data, mime_type=f"image/{ext}"))

        parts.append(types.Part.from_text(text=query))
        msgs.append(
            types.Content(
                role="user",
                parts=parts,
            )
        )

        MAX_TOKENS = min(self.config.max_tokens, self.context_length)
        MAX_OUTPUT_TOKENS = min(
            MAX_TOKENS, self.max_output_tokens, self.config.max_output_tokens
        )
        prompt_token_count = self.calculate_token_count(
            self.model_name,
            self.system_prompt,
            msgs,
            imgs=None if filepath is None else [filepath],
        )
        max_output_tokens = min(
            MAX_OUTPUT_TOKENS,
            self.context_length - prompt_token_count,
        )

        # accumulated_token_count = 0  # Accumulated token count across iterations
        config = self.custom_config(max_output_tokens, response_mode, response_format)
        try:
            client = genai.Client(api_key=os.environ["GEMINI_API_KEY"])
            response = client.models.generate_content(
                model=self.model_name,
                contents=msgs,  # type: ignore
                config=config,
            )

            usage = response.usage_metadata
            if usage:
                logger.info("Usage: %s", usage)

            response_text = getattr(response, "text", None)
            if response_text is None:
                raise RuntimeError("response.text is None")

            if response_mode == ResponseMode.SO:
                custom_format = response.parsed
                custom_dict = custom_format.model_dump()  # type: ignore
                for k, v in custom_dict.items():
                    logger.info("%s : %s", k, v)

            output.append(
                {"role": CreatorRole.ASSISTANT.value, "content": response_text}
            )
            return output
        except Exception as e:
            logger.error("Exception: %s", e)
            raise

    async def run_async(
        self, query: str, context: list[MessageBlock | dict] | None, **kwargs
    ) -> list[MessageBlock | dict]:
        """
        Synchronously run the LLM model with the given query and context.

        Args:
            query (str): The query to be processed by the LLM model.
            context (list[MessageBlock | dict] | None): The context to be used for the LLM model.
            **kwargs: Additional keyword arguments.
                * `filepath` (str | None): Path to the image file.
                * `mode` (ResponseMode | None): Ouput mode.
                * `format` (BaseModel | None): Output structure.

        Returns:
            list[MessageBlock | dict]: The list of messages generated by the LLM model.

        Notes:
        * Single-Turn Execution.
        """
        response_mode: Optional[ResponseMode] = kwargs.get("mode", ResponseMode.DEFAULT)
        response_format: Optional[Type[T]] = kwargs.get("format")  # type: ignore
        self.validate(response_mode, response_format)  # Raise an exception if invalid

        msgs: list[types.Content] = []
        output: list[MessageBlock | dict] = []

        if context:
            for ctx in context:
                msgs.append(
                    types.Content(
                        role=ctx["role"],
                        parts=[types.Part.from_text(text=ctx["content"])],
                    )
                )
        parts = []
        filepath: str | None = kwargs.get("filepath", None)
        if filepath:
            with open(filepath, "rb") as f:
                data = f.read()
            ext = os.path.splitext(filepath)[-1][1:]
            ext = "jpeg" if ext == "jpg" else ext
            parts.append(types.Part.from_bytes(data=data, mime_type=f"image/{ext}"))

        parts.append(types.Part.from_text(text=query))
        msgs.append(
            types.Content(
                role="user",
                parts=parts,
            )
        )
        MAX_TOKENS = min(self.config.max_tokens, self.context_length)
        MAX_OUTPUT_TOKENS = min(
            MAX_TOKENS, self.max_output_tokens, self.config.max_output_tokens
        )
        prompt_token_count = self.calculate_token_count(
            self.model_name,
            self.system_prompt,
            msgs,
            imgs=None if filepath is None else [filepath],
        )
        max_output_tokens = min(
            MAX_OUTPUT_TOKENS,
            self.context_length - prompt_token_count,
        )

        # accumulated_token_count = 0  # Accumulated token count across iterations
        config = self.custom_config(max_output_tokens, response_mode, response_format)
        try:
            client = genai.Client(api_key=os.environ["GEMINI_API_KEY"])
            response = client.models.generate_content(
                model=self.model_name,
                contents=msgs,  # type: ignore
                config=config,
            )

            usage = response.usage_metadata
            if usage:
                logger.info("Usage: %s", usage)

            response_text = getattr(response, "text", None)
            if response_text is None:
                raise RuntimeError("response.text is None")

            if response_mode == ResponseMode.SO:
                custom_format = response.parsed
                custom_dict = custom_format.model_dump()  # type: ignore
                for k, v in custom_dict.items():
                    logger.info("%s : %s", k, v)

            output.append(
                {"role": CreatorRole.ASSISTANT.value, "content": response_text}
            )
            return output
        except Exception as e:
            logger.error("Exception: %s", e)
            raise

    def interpret(
        self,
        query: str,
        context: list[MessageBlock | dict[str, Any]] | None,
        filepath: str,
        **kwargs,
    ) -> list[MessageBlock | dict[str, Any]]:
        return self.run(query=query, context=context, filepath=filepath, **kwargs)

    async def interpret_async(
        self,
        query: str,
        context: list[MessageBlock | dict[str, Any]] | None,
        filepath: str,
        **kwargs,
    ) -> list[MessageBlock | dict[str, Any]]:
        return await self.run_async(
            query=query, context=context, filepath=filepath, **kwargs
        )
